#include <cstdio>
#include <string>
#include <list>
#include <set>
#include <map>
#include <argh.hh>

#include "autoptr"
#include "riffhead.hh"

using namespace std;

#define bufsize 512

bool quiet = false;
static bool filepos=false;

static set<string> listignores;

struct tieto
{
    int ind;
    string field, info;
    long startpos;
    long bytes, overuse, excess;
    
    const string getstring() const
    {
        const string &k = field;
        
        char BufO[64]=""; if(overuse)sprintf(BufO, ", %ld OVERUSE", overuse);
        char BufE[64]=""; if(excess)sprintf(BufE, ", %ld data", excess);
        
        char Buf[512];
        string result(Buf, sprintf(Buf, "%*s", ind,""));
        if(filepos)
            result += string(Buf, sprintf(Buf, "%ld: ", startpos));
        result += k;
        result += string(Buf,
          sprintf(Buf, " (%ld bytes%s%s)",
            bytes,
            BufO,
            BufE));
        if(info.size()) { result += " ("; result += info; result += ')'; }
        return result;
    }
};

static list<tieto> tiedot;

unsigned ReadDW(FILE *fp)
{
    char Buf[4];
    fread(Buf, 1, 4, fp);
    return ((unsigned)((unsigned char)Buf[0]))
         | ((unsigned)((unsigned char)Buf[1]) << 8)
         | ((unsigned)((unsigned char)Buf[2]) << 16)
         | ((unsigned)((unsigned char)Buf[3]) << 24);
}
unsigned ReadW(FILE *fp)
{
    char Buf[2];
    fread(Buf, 1, 2, fp);
    return ((unsigned)((unsigned char)Buf[0]))
         | ((unsigned)((unsigned char)Buf[1]) << 8);
}

class readermap: public map<string, autoptr<RiffHead> >, public ptrable {};
static readermap *readers = NULL;
static class poo{public:~poo(){delete readers;}} poo;

RiffHead::~RiffHead() {}
void RegisterFormat(const char *code, autoptr<RiffHead> instance)
{
    if(!readers)readers = new readermap;
    (*readers)[code] = instance;
}

void FindSomething(FILE *fp)
{
    static int ind=-1;struct ipp{int i;ipp():i(ind){ind+=1;}~ipp(){ind=i;}}ipp;
    
    if(ferror(fp))
    {
        perror("");
        return;
    }
    
    tieto data;
    data.startpos = ftell(fp);
    data.ind = ind;
    
    char Buf[5] = "????";
    fread(Buf, 1, 4, fp);
    if(feof(fp))return;

    data.bytes = ReadDW(fp);

    data.field = Buf;

    long pos = ftell(fp);
    long endpos = pos + data.bytes;
    
    unsigned skip = (data.bytes + 1) & (~1UL);
    pos += skip;
    if(pos & 1) ++pos;
    
    list<tieto>::iterator between = tiedot.end(); --between;
    
    
    if(data.field == "LIST"
    || data.field == "RIFF")
    {
        char Buf2[5];
        fread(Buf2, 1, 4, fp);
        Buf2[4] = 0;
        string s2 = Buf2;
        
        data.field += '(';
        data.field += Buf2;
        data.field += ')';
        
        if(listignores.find(s2) == listignores.end())
            while(!feof(fp) && !ferror(fp) && ftell(fp) < endpos)
            {
                if(!quiet)
                {
                    double pos = ftell(fp) * 100.0 / endpos;
                    sprintf(Buf2, "%4.1f", pos);
                    static string prosk;
                    if(prosk!=Buf2) {prosk=Buf2;fprintf(stderr,"%s%%\r",Buf2);fflush(stderr);}
                }
                FindSomething(fp);
            }
    }
    else
    {
        map<string, autoptr<RiffHead> >::iterator i = readers->find(data.field);
        if(i != readers->end())
        {
            data.info = i->second->Read(fp, endpos-ftell(fp));
        }
    }
    
    long pos2 = ftell(fp);
    data.overuse = pos2 > endpos ? pos2-endpos : 0;
    data.excess  = pos2 < endpos ? endpos-pos2 : 0;
    
    if(data.overuse > 0)
        fseek(fp, pos, SEEK_SET);
    else if(data.excess > 0) // bufsize-8)
        fseek(fp, pos, SEEK_SET);
    /*else if(excess > 0)
    {
        char Buf[bufsize];
        fread(Buf, 1, excess, fp);
    }*/

    tiedot.insert(++between, data);
}

#include <unistd.h>
#include <cstdlib>

static void Handle(FILE *fp)
{
    if(!readers)readers = new readermap;
    
    {tieto tmp; tiedot.push_back(tmp);}
    
    {static char Buf[bufsize];setbuffer(fp,Buf,bufsize);}
    //setbuffer(fp, NULL, 64);
    
    while(!feof(fp))
        FindSomething(fp);
    
    tiedot.erase(tiedot.begin());
    
    list<tieto>::const_iterator i;
    for(i=tiedot.begin(); i!=tiedot.end(); ++i)
    {
        const string &s = i->getstring();
        fwrite(s.data(), 1, s.size(), stdout);
        putchar('\n');
    }
}

int main(int argc, const char *const *argv)
{
    list<string> files;
    
    ParamHandler Argh;
    Argh.AddLong("help",       'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version",    'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("quiet",      'q').SetBool().SetDesc("Disable displaying % of completion.");
    Argh.AddLong("listignore", 'n').SetString(4,4).SetDesc(
      "Ignore LIST chunks of specified type,\n"
      "Example: -nmovi", "<type>");
    Argh.AddLong("filepos",    'p').SetBool().SetDesc("Display filepos");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "This is riffdump-"VERSION"\n"
                    "Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n"
                    "\n"
                    "Displays the structure of a RIFF file (AVI, WAV, ...)\n"
                    "\n"
                    "Usage: riffdump [<option> [<...>]] <file>\n"
                    "\n"
                    "Options:\n");
                Argh.ListOptions();
                printf("\n"
                       "riffdump is a partner program of animmaker.\n"
                       "No warranty whatsoever.\n");
                return 0;
            case 'n':
                listignores.insert(Argh.GetString());
                break;
            case 'q':
                quiet = Argh.GetBool();
                break;
            case 'p':
                filepos = Argh.GetBool();
                break;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    
    int errs = 0;
    
    if(!files.size())
    {
        if(isatty(STDIN_FILENO))
        {
            fprintf(stderr, "riffdump: AVI data not written to a terminal.\n");
            ++errs;
        }
        else
            Handle(stdin);
    }
    else
    {
        list<string>::const_iterator i;
        for(i=files.begin(); i!=files.end(); ++i)
        {
            if(*i == "-")
            {
                if(isatty(STDIN_FILENO))
                {
                    fprintf(stderr, "riffdump: AVI data not written to a terminal.\n");
                    ++errs;
                }
                else
                    Handle(stdin);
                continue;
            }
            FILE *fp = fopen(i->c_str(), "rb");
            if(!fp)
                { perror(i->c_str()); ++errs; continue; }
            Handle(fp);
            fclose(fp);
        }
    }
    return errs;
}
